#define __PROG_TYPES_COMPAT__

#define F_CPU			16000000

#include <avr/pgmspace.h>
#include <avr/io.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>

#include "lcd.h"
#include "gfx.h"
#include "utils.h"

// each shape is made of 4 squares
struct square {
	int16_t x, y;
	uint16_t width;
	uint16_t color;
	uint16_t fill;
};

// shape info
struct shape {
	struct square elems[4];
	int16_t total_width, total_height;
	uint8_t rotation, type, ok;
	uint16_t x_max, x_min, y_max, y_min;
};

// static variables
static int8_t lost = 0, game_over = 0;
static int8_t last_button = 0;
static struct shape shapes[SHAPES_NO];
static char matrix[DISPLAY_WIDTH][DISPLAY_HEIGHT];

struct shape R, L, T, S;
unsigned int i, j, k;

// function that sets buttons
static void button_set(uint8_t button)
{
	BUTTON_REG &= ~(1 << button);
	BUTTON_PORT |= (1 << button);
}

// function that detects whether a button is pressed
uint8_t button_press(uint8_t button)
{
	if (!(BUTTON_PIN & (1 << button))) {
		_delay_ms(50);
		
		if (!(BUTTON_PIN & (1 << button))) {
			_delay_ms(50);
			
			if (!(BUTTON_PIN & (1 << button)))
				return 1;
		}
	}
	
	return 0;
}

// writes Welcome! at the beginning
static void write_welcome(int16_t x, int16_t y)
{
	GFX_set_text_size(2);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x, y);
	GFX_print_str("Welcome!");
}

// writes Goodbye! after Exit selected
static void write_goodbye(int16_t x, int16_t y)
{
	GFX_set_text_size(2);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x, y);
	GFX_print_str("Goodbye!");
}

// writes Game over! at the end of a game
static void write_gameover(int16_t x, int16_t y)
{
	GFX_set_text_size(2);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x, y);
	GFX_print_str("Game over!");
}


// displays the user's menu
static void write_menu(int16_t x, int16_t y)
{
	char games_lost[10];

	GFX_set_text_size(2);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(0, 0);
	GFX_print_str("Tetris");

	GFX_set_text_size(1);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x, y + 10);
	GFX_print_str("Play");

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x, y + 20);
	GFX_print_str("Exit");

	sprintf(games_lost, "Lost:  %d", lost);
	GFX_set_text_color(ST7735_RED);
	GFX_set_cursor(x, y + 40);
	GFX_print_str(games_lost);
}

// initializes shapes and saves them in the shapes array
static void init_shapes()
{
	for (i = 0; i < SHAPES_NO; i++) {
		R.elems[i].width = L.elems[i].width = T.elems[i].width = S.elems[i].width = SQUARE_DIM;
		R.elems[i].fill = L.elems[i].fill = T.elems[i].fill = S.elems[i].fill = FILLED;

		R.elems[i].color = ST7735_CYAN;
		L.elems[i].color = ST7735_YELLOW;
		T.elems[i].color = ST7735_MAGENTA;
		S.elems[i].color = ST7735_BLUE;
	}

	R.elems[0].x = R.elems[1].x = R.elems[2].x = R.elems[3].x = X_MIN_UP;
	R.elems[0].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 + 2 * SQUARE_DIM; //103
	R.elems[1].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 + SQUARE_DIM; //96
	R.elems[2].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2; //89
	R.elems[3].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 - SQUARE_DIM; //82
	R.total_width = 4 * SQUARE_DIM;
	R.total_height = SQUARE_DIM;
	R.rotation = 0;
	R.type = 0;
	R.x_min = X_MIN_UP;
	R.x_max = X_MIN_UP + R.total_height;
	R.y_min = R.elems[0].y - R.total_width;
	R.y_max = R.elems[0].y;

	L.elems[0].x = X_MIN_UP;
	L.elems[1].x = L.elems[2].x = L.elems[3].x = X_MIN_UP + SQUARE_DIM;
	L.elems[0].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 - SQUARE_DIM;
	L.elems[1].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 + SQUARE_DIM;
	L.elems[2].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2;
	L.elems[3].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 - SQUARE_DIM;
	L.total_width = 3 * SQUARE_DIM;
	L.total_height = 2 * SQUARE_DIM;
	L.rotation = 0;
	L.type = 1;
	L.x_min = X_MIN_UP;
	L.x_max = X_MIN_UP + L.total_height;
	L.y_min = L.elems[1].y - L.total_width;
	L.y_max = L.elems[1].y;

	T.elems[0].x = X_MIN_UP;
	T.elems[1].x = T.elems[2].x = T.elems[3].x = X_MIN_UP + SQUARE_DIM;
	T.elems[0].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2;
	T.elems[1].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 + SQUARE_DIM;
	T.elems[2].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2;
	T.elems[3].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 - SQUARE_DIM;
	T.total_width = 3 * SQUARE_DIM;
	T.total_height = 2 * SQUARE_DIM;
	T.rotation = 0;
	T.type = 2;
	T.x_min = X_MIN_UP;
	T.x_max = X_MIN_UP + T.total_height;
	T.y_min = T.elems[1].y - T.total_width;
	T.y_max = T.elems[1].y;

	S.elems[0].x = S.elems[1].x = X_MIN_UP;
	S.elems[2].x = S.elems[3].x = X_MIN_UP + SQUARE_DIM;
	S.elems[0].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 + SQUARE_DIM;
	S.elems[1].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2;
	S.elems[2].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 + SQUARE_DIM;
	S.elems[3].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2;
	S.total_width = 2 * SQUARE_DIM;
	S.total_height = 2 * SQUARE_DIM;
	S.rotation = 0;
	S.type = 3;
	S.x_min = X_MIN_UP;
	S.x_max = X_MIN_UP + S.total_height;
	S.y_min = S.elems[0].y - S.total_width;
	S.y_max = S.elems[0].y;

	shapes[0] = R;
	shapes[1] = L;
	shapes[2] = T;
	shapes[3] = S;
}


// draws a given shape, taking into consideration its button-caused modifications
static void draw_shape(struct shape s)
{
	// do not draw any more shapes if the game is over
	if (game_over == 1)
		return;

	// draw the initial shape
	for (i = 0; i < 4; i++) {
		GFX_draw_rect(s.elems[i].x,
					  s.elems[i].y, 
					  s.elems[i].width, 
					  s.elems[i].width, 
					  s.elems[i].color, 
					  s.elems[i].fill);
		matrix[(s.elems[i].x - 1) / SQUARE_DIM][(s.elems[i].y - Y_OFFSET - 1) / SQUARE_DIM] = 1;
	}

	// the shape will be falling
	s.ok = 0;

	// while the shape still has space to fall and the game is not over
	while (s.ok == 0 && game_over != 1) {

		// check if it can be drawn at the next step
		for (i = 0; i < 4; i++)
			if (matrix[(s.x_max + SQUARE_DIM - 1) / SQUARE_DIM][(s.elems[i].y - Y_OFFSET - 1) / SQUARE_DIM] == 1) {
				s.ok = 1;
				break;
			}

		// make the previous shape disappear
		for (i = 0; i < 4; i++) {
			GFX_draw_rect(s.elems[i].x,
						  s.elems[i].y, 
					  	  s.elems[i].width, 
					  	  s.elems[i].width, 
					  	  ST7735_BLACK, 
					  	  s.elems[i].fill);
			matrix[(s.elems[i].x - 1) / SQUARE_DIM][(s.elems[i].y - Y_OFFSET - 1) / SQUARE_DIM] = 0;
		}

		// apply rotations if buttons are pressed
		if (s.rotation % 4 == 0) {
			if (s.type == 0) {
				for (i = 1; i < 4; i++) {
					s.elems[i].y = s.elems[0].y - i * SQUARE_DIM;
					s.elems[i].x = s.elems[0].x;
				}
				s.total_width = 4 * SQUARE_DIM;
				s.total_height = SQUARE_DIM;
				s.x_min = s.elems[0].x;
				s.x_max = s.elems[0].x + s.total_height;
				s.y_min = s.elems[0].y - s.total_width;
				s.y_max = s.elems[0].y;
			}
			if (s.type == 1) {
				s.elems[1].x = s.elems[2].x = s.elems[3].x = s.elems[0].x + SQUARE_DIM;
				s.elems[3].y = s.elems[0].y;
				s.elems[1].y = s.elems[0].y + 2 * SQUARE_DIM;
				s.elems[2].y = s.elems[0].y + SQUARE_DIM;
				s.total_width = 3 * SQUARE_DIM;
				s.total_height = 2 * SQUARE_DIM;
				s.x_min = s.elems[0].x;
				s.x_max = s.elems[0].x + s.total_height;
				s.y_min = s.elems[1].y - s.total_width;
				s.y_max = s.elems[1].y;
			}
			if (s.type == 2) {
				s.elems[1].x = s.elems[2].x = s.elems[3].x = s.elems[0].x + SQUARE_DIM;
				s.elems[1].y = s.elems[0].y + SQUARE_DIM;
				s.elems[2].y = s.elems[0].y;
				s.elems[3].y = s.elems[0].y - SQUARE_DIM;
				s.total_width = 3 * SQUARE_DIM;
				s.total_height = 2 * SQUARE_DIM;
				s.x_min = s.elems[0].x;
				s.x_max = s.elems[0].x + s.total_height;
				s.y_min = s.elems[1].y - s.total_width;
				s.y_max = s.elems[1].y;
			}
		}
		if (s.rotation % 4 == 1 || s.rotation % 4 == -3) {
			if (s.type == 0) {
				for (i = 1; i < 4; i++) {
					s.elems[i].x = s.elems[0].x + i * SQUARE_DIM;
					s.elems[i].y = s.elems[0].y;
				}
				s.total_width = SQUARE_DIM;
				s.total_height = 4 * SQUARE_DIM;
				s.x_min = s.elems[0].x;
				s.x_max = s.elems[0].x + s.total_height;
				s.y_min = s.elems[0].y - s.total_width;
				s.y_max = s.elems[0].y;
			}
			if (s.type == 1) {
				s.elems[0].x = s.elems[3].x;
				s.elems[1].x = s.elems[3].x - 2 * SQUARE_DIM;
				s.elems[2].x = s.elems[3].x - SQUARE_DIM;
				s.elems[1].y = s.elems[2].y = s.elems[3].y;
				s.elems[0].y = s.elems[3].y - SQUARE_DIM;
				s.total_width = 2 * SQUARE_DIM;
				s.total_height = 3 * SQUARE_DIM;
				s.x_min = s.elems[1].x;
				s.x_max = s.elems[1].x + s.total_height;
				s.y_min = s.elems[1].y - s.total_width;
				s.y_max = s.elems[1].y;
			}
			if (s.type == 2) {
				s.elems[1].y = s.elems[3].y = s.elems[2].y;
				s.elems[0].y = s.elems[2].y - SQUARE_DIM;
				s.elems[0].x = s.elems[2].x;
				s.elems[1].x = s.elems[2].x - SQUARE_DIM;
				s.elems[3].x = s.elems[2].x + SQUARE_DIM;
				s.total_width = 2 * SQUARE_DIM;
				s.total_height = 3 * SQUARE_DIM;
				s.x_min = s.elems[1].x;
				s.x_max = s.elems[1].x + s.total_height;
				s.y_min = s.elems[1].y - s.total_width;
				s.y_max = s.elems[1].y;
			}
		}
		if (s.rotation % 4 == 2 || s.rotation % 4 == -2) {
			if (s.type == 0) {
				for (i = 1; i < 4; i++) {
					s.elems[i].y = s.elems[0].y + i * SQUARE_DIM;
					s.elems[i].x = s.elems[0].x;
				}
				s.total_width = 4 * SQUARE_DIM;
				s.total_height = SQUARE_DIM;
				s.x_min = s.elems[0].x;
				s.x_max = s.elems[0].x + s.total_height;
				s.y_min = s.elems[0].y - s.total_width;
				s.y_max = s.elems[0].y;
			}
			if (s.type == 1) {
				s.elems[1].x = s.elems[2].x = s.elems[3].x;
				s.elems[0].x = s.elems[3].x + SQUARE_DIM;
				s.elems[0].y = s.elems[3].y;
				s.elems[2].y = s.elems[3].y - SQUARE_DIM;
				s.elems[1].y = s.elems[3].y - 2 * SQUARE_DIM;
				s.total_width = 3 * SQUARE_DIM;
				s.total_height = 2 * SQUARE_DIM;
				s.x_min = s.elems[3].x;
				s.x_max = s.elems[3].x + s.total_height;
				s.y_min = s.elems[3].y - s.total_width;
				s.y_max = s.elems[3].y;
			}
			if (s.type == 2) {
				s.elems[1].x = s.elems[3].x = s.elems[2].x;
				s.elems[0].x = s.elems[2].x + SQUARE_DIM;
				s.elems[1].y = s.elems[2].y - SQUARE_DIM;
				s.elems[3].y = s.elems[2].y + SQUARE_DIM;
				s.elems[0].y = s.elems[2].y;
				s.total_width = 3 * SQUARE_DIM;
				s.total_height = 2 * SQUARE_DIM;
				s.x_min = s.elems[1].x;
				s.x_max = s.elems[1].x + s.total_height;
				s.y_min = s.elems[3].y - s.total_width;
				s.y_max = s.elems[3].y;
			}
		}
		if (s.rotation % 4 == 3 || s.rotation % 4 == -1) {
			if (s.type == 0) {
				for (i = 1; i < 4; i++) {
					s.elems[i].x = s.elems[0].x - i * SQUARE_DIM;
					s.elems[i].y = s.elems[0].y;
				}
				s.total_width = SQUARE_DIM;
				s.total_height = 4 * SQUARE_DIM;
				s.x_min = s.elems[0].x;
				s.x_max = s.elems[0].x + s.total_height;
				s.y_min = s.elems[0].y - s.total_width;
				s.y_max = s.elems[0].y;
			}
			if (s.type == 1) {
				s.elems[0].x = s.elems[3].x;
				s.elems[1].x = s.elems[3].x + 2 * SQUARE_DIM;
				s.elems[2].x = s.elems[3].x + SQUARE_DIM;
				s.elems[1].y = s.elems[2].y = s.elems[3].y;
				s.elems[0].y = s.elems[3].y + SQUARE_DIM;
				s.total_width = 2 * SQUARE_DIM;
				s.total_height = 3 * SQUARE_DIM;
				s.x_min = s.elems[3].x;
				s.x_max = s.elems[3].x + s.total_height;
				s.y_min = s.elems[0].y - s.total_width;
				s.y_max = s.elems[0].y;
			}
			if (s.type == 2) {
				s.elems[1].y = s.elems[3].y = s.elems[2].y;
				s.elems[0].y = s.elems[2].y + SQUARE_DIM;
				s.elems[0].x = s.elems[2].x;
				s.elems[1].x = s.elems[2].x + SQUARE_DIM;
				s.elems[3].x = s.elems[2].x - SQUARE_DIM;
				s.total_width = 2 * SQUARE_DIM;
				s.total_height = 3 * SQUARE_DIM;
				s.x_min = s.elems[3].x;
				s.x_max = s.elems[3].x + s.total_height;
				s.y_min = s.elems[0].y - s.total_width;
				s.y_max = s.elems[0].y;
			}
		}

		// draw the next step shape
		for (i = 0; i < 4; i++) {
			GFX_draw_rect(s.elems[i].x + SQUARE_DIM,
					  	  s.elems[i].y,
					  	  s.elems[i].width, 
					  	  s.elems[i].width, 
					  	  s.elems[i].color, 
					  	  s.elems[i].fill);
			matrix[(s.elems[i].x + SQUARE_DIM - 1) / SQUARE_DIM][(s.elems[i].y - Y_OFFSET - 1) / SQUARE_DIM] = 1;
		}

		_delay_ms(250);

		// move the shape to the left
		if (button_press(BUTTON_LEFT)) {
			// make the previous shape disappear
			for (i = 0; i < 4; i++) {
				GFX_draw_rect(s.elems[i].x + SQUARE_DIM,
						  	  s.elems[i].y,
						  	  s.elems[i].width, 
						  	  s.elems[i].width, 
						  	  ST7735_BLACK, 
						  	  s.elems[i].fill);
				// mark the positions as free
				matrix[(s.elems[i].x + SQUARE_DIM - 1) / SQUARE_DIM][(s.elems[i].y - Y_OFFSET - 1) / SQUARE_DIM] = 0;
			}
			// draw the new shape
			for (i = 0; i < 4; i++) {
				GFX_draw_rect(s.elems[i].x + SQUARE_DIM,
						  	  s.elems[i].y + SQUARE_DIM,
						  	  s.elems[i].width, 
						  	  s.elems[i].width, 
						  	  s.elems[i].color, 
						  	  s.elems[i].fill);
				// mark the positions as occupied
				matrix[(s.elems[i].x + SQUARE_DIM - 1) / SQUARE_DIM][(s.elems[i].y + SQUARE_DIM - Y_OFFSET - 1) / SQUARE_DIM] = 1;
			}
			// modify shape variables
			for (i = 0; i < 4; i++) {
				s.elems[i].y += SQUARE_DIM;
			}
			s.y_max += SQUARE_DIM;
			s.y_min += SQUARE_DIM;
		}
		// move the shape to the right
		if (button_press(BUTTON_RIGHT)) {
			// make the previous shape disappear
			for (i = 0; i < 4; i++) {
				GFX_draw_rect(s.elems[i].x + SQUARE_DIM,
						  	  s.elems[i].y,
						  	  s.elems[i].width, 
						  	  s.elems[i].width, 
						  	  ST7735_BLACK, 
						  	  s.elems[i].fill);
				// mark the positions as free
				matrix[(s.elems[i].x + SQUARE_DIM - 1) / SQUARE_DIM][(s.elems[i].y - Y_OFFSET - 1) / SQUARE_DIM] = 0;
			}
			// draw the new shape
			for (i = 0; i < 4; i++) {
				GFX_draw_rect(s.elems[i].x + SQUARE_DIM,
						  	  s.elems[i].y - SQUARE_DIM,
						  	  s.elems[i].width, 
						  	  s.elems[i].width, 
						  	  s.elems[i].color, 
						  	  s.elems[i].fill);
				// mark the positions as occupied
				matrix[(s.elems[i].x + SQUARE_DIM - 1) / SQUARE_DIM][(s.elems[i].y - SQUARE_DIM - Y_OFFSET - 1) / SQUARE_DIM] = 1;
			}
			// modify shape variables
			for (i = 0; i < 4; i++) {
				s.elems[i].y -= SQUARE_DIM;
			}
			s.y_max -= SQUARE_DIM;
			s.y_min -= SQUARE_DIM;
		}

		// detect rotation modifications
		if (button_press(BUTTON_UP))
			s.rotation++;
		if (button_press(BUTTON_DOWN))
			s.rotation--;

		// modify shape variables
		for (i = 0; i < 4; i++) {
			s.elems[i].x += SQUARE_DIM;
		}
		s.x_max += SQUARE_DIM;
		s.x_min += SQUARE_DIM;

		// hitting the bottom edge
		if (s.x_max >= X_MAX_DOWN) {
			s.x_max = X_MAX_DOWN;
			s.ok = 1;
			break;
		}		

	}

	// detect game end (no more shapes have enough space)
	if (s.x_min <= X_MIN_UP + SQUARE_DIM) {
		game_over = 1;
		lost++;
		return;
	}
}

int main(void)
{	
	// set the buttons
	button_set(BUTTON_UP);
	button_set(BUTTON_DOWN);
	button_set(BUTTON_LEFT);
	button_set(BUTTON_RIGHT);
	button_set(BUTTON_OK);
	last_button = UP;

	// initialize the shapes and the matrix
	init_shapes();
	for (i = 0; i < DISPLAY_HEIGHT; i++)
		for (j = 0; j < DISPLAY_WIDTH; j++)
			matrix[i][j] = 0;

	// initialize LCD
	LCD_init();
	GFX_init(128, 160);
	LCD_fill_screen(ST7735_BLACK);
	write_welcome(0, 0);
	_delay_ms(2000);
	LCD_fill_screen(ST7735_BLACK);

	while (1) {

		// draw the menu (Play selected by default)
		write_menu(5, 30);
		if (last_button == UP) {
			last_button = UP;
			GFX_draw_rect(38, 129, 10, 28, ST7735_CYAN, 0);
		}

		// move the rectangle used for selection
		if (button_press(BUTTON_DOWN) == 1) {
			GFX_draw_rect(38, 129, 10, 28, ST7735_BLACK, 0);
			last_button = DOWN;
			GFX_draw_rect(48, 129, 10, 28, ST7735_CYAN, 0);
		}
		else if (button_press(BUTTON_UP) == 1) {
			GFX_draw_rect(48, 129, 10, 28, ST7735_BLACK, 0);
			last_button = UP;
			GFX_draw_rect(38, 129, 10, 28, ST7735_CYAN, 0);
		}

		// Play selected
		if (last_button == UP && button_press(BUTTON_OK) == 1) {

			//y are valori valide intre 25 si 153 (creste de la dreapta la stanga display-ului)
			//x are valori valide intre 0 si 128 (creste de sus in jos)

			// begin the game
			LCD_fill_screen(ST7735_BLACK);
			_delay_ms(1000);

			// continuously draw "random" shapes
			while (game_over != 1) {
				uint8_t randNumber;
				randNumber = (uint8_t)rand();
				draw_shape(shapes[randNumber % 4]);
			}

			if (game_over == 1) {
				// refresh variables
				game_over = 0;
				last_button = UP;
				init_shapes();
				for (i = 0; i < DISPLAY_HEIGHT; i++)
					for (j = 0; j < DISPLAY_WIDTH; j++)
						matrix[i][j] = 0;

				// manage display (make it friendly)
				LCD_fill_screen(ST7735_BLACK);
				write_gameover(0, 0);
				_delay_ms(2000);
				LCD_fill_screen(ST7735_BLACK);
				write_menu(5, 30);
				if (last_button == UP) {
					last_button = UP;
					GFX_draw_rect(38, 129, 10, 28, ST7735_CYAN, 0);
				}
				if (button_press(BUTTON_DOWN) == 1) {
					GFX_draw_rect(38, 129, 10, 28, ST7735_BLACK, 0);
					last_button = DOWN;
					GFX_draw_rect(48, 129, 10, 28, ST7735_CYAN, 0);
				}
				else if (button_press(BUTTON_UP) == 1) {
					GFX_draw_rect(48, 129, 10, 28, ST7735_BLACK, 0);
					last_button = UP;
					GFX_draw_rect(38, 129, 10, 28, ST7735_CYAN, 0);
				}
			}

		}

		// Exit selected
		if (last_button == DOWN && button_press(BUTTON_OK) == 1) {
			last_button = OK;
			LCD_fill_screen(ST7735_BLACK);
			write_goodbye(0, 0);
			_delay_ms(2000);
			LCD_fill_screen(ST7735_BLACK);
			LCD_end();
			break;
		}

		_delay_ms(50);
	}

	return 0;
}
